/*
 * Copyright (c) 2012 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *					Anthony Mallet on Wed Feb  8 2012
 */
#include "acgenomix.h"

#include <err.h>
#include <pthread.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>


#include <tcl.h>

#include "genomix.h"


/* --- local data ---------------------------------------------------------- */

/* list of extra package paths */
static const char *sysdir[] = {
  GENOMIX_SYSDIR,
  "."
};

/* tcl async handler for signals */
static Tcl_AsyncHandler async;
static pthread_t asyncthread;
static int asyncsig;


/* --- signal handlers ----------------------------------------------------- */

static void
sighandler(int sig)
{
  /* make sure to send the signal to the lwp blocked on select() (tcl vwait
   * command). It is not clear whether pthread_kill is async signal safe? */
  if (!pthread_equal(pthread_self(), asyncthread)) {
    pthread_kill(asyncthread, sig);
    return;
  }

  asyncsig = sig;
  Tcl_AsyncMark(async);
}

static int
asyncsighandler(ClientData clientData, Tcl_Interp *interp, int code)
{
  /* interp can be NULL: use clientData which points to the actual interp */
  Tcl_SetVar(clientData, "forever", strsignal(asyncsig), TCL_GLOBAL_ONLY);
  return code;
}


/* --- main ---------------------------------------------------------------- */

/** \brief startup...
 */
int
main(int argc, const char *argv[])
{
  struct sigaction sact;
  Tcl_Interp *interp;
  const char *v;
  Tcl_Obj *obj;
  char *args;
  int p, s;

  /* initialize event watcher */
  gnmx_watch_init();

  /* create tcl interpreter */
  interp = Tcl_CreateInterp();
  if (!interp) {
    warnx("cannot create Tcl interpreter");
    return 2;
  }

  /* make arguments available in argc/argv variables */
  obj = Tcl_NewIntObj(argc - 1);
  Tcl_SetVar2Ex(interp, "argc", NULL, obj, TCL_GLOBAL_ONLY);

  args = Tcl_Merge(argc-1, argv+1);
  Tcl_SetVar(interp, "argv", args, TCL_GLOBAL_ONLY);
  Tcl_Free(args);

  Tcl_SetVar(interp, "argv0", argv[0], TCL_GLOBAL_ONLY);

  /* set the "tcl_interactive" variable. */
  Tcl_SetVar(interp, "tcl_interactive", "0", TCL_GLOBAL_ONLY);

  /* initialize interpreter */
  if (Tcl_Init(interp) == TCL_ERROR) goto error;
  if (!Tcl_PkgRequire(interp, "Tcl", "8.5", 0))
    goto error;

  /* create signal handlers */
  async = Tcl_AsyncCreate(asyncsighandler, interp);
  asyncthread = pthread_self();

  sact.sa_handler = sighandler;
  sigemptyset(&sact.sa_mask);
  sact.sa_flags = 0;

  s = sigaction(SIGINT, &sact, NULL);
  s = sigaction(SIGQUIT, &sact, NULL);
  s = sigaction(SIGTERM, &sact, NULL);

  /* create genomix commands */
  Tcl_CreateObjCommand(interp, "genomix::rpath", gnmx_rpath, NULL, NULL);
  Tcl_CreateObjCommand(interp, "genomix::load", gnmx_load, NULL, NULL);

  /* set path to packages */
  obj = Tcl_GetVar2Ex(interp, "auto_path", NULL, TCL_GLOBAL_ONLY);
  for(p=0; p<sizeof(sysdir)/sizeof(sysdir[0]); p++) {
    s = Tcl_ListObjAppendElement(interp, obj, Tcl_NewStringObj(sysdir[p], -1));
    if (s != TCL_OK) goto error;
  }
  Tcl_SetVar2Ex(interp, "auto_path", NULL, obj, TCL_GLOBAL_ONLY);

  /* require daemon package */
  if (!Tcl_PkgRequire(interp, "genomixd", PACKAGE_VERSION, 1))
    goto error;

  s = Tcl_Eval(interp, "main");
  if (s != TCL_OK) goto error;

  Tcl_DeleteInterp(interp);

  /* run any local atexit that could not be registered via the real atexit(),
   * normally none. */
  gnmx_atexit_invoke(NULL);
  return 0;

error:
  v = Tcl_GetVar(interp, "verbose", TCL_GLOBAL_ONLY);
  if (v && atoi(v))
    warnx("%s", Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY));
  else
    warnx("%s", Tcl_GetStringResult(interp));
  Tcl_DeleteInterp(interp);
  gnmx_atexit_invoke(NULL);
  return 2;
}
