#
# Copyright (c) 2012-2013 LAAS/CNRS
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software for any purpose
# with or without   fee is hereby granted, provided   that the above  copyright
# notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS  SOFTWARE INCLUDING ALL  IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS. IN NO EVENT SHALL THE AUTHOR  BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR  ANY DAMAGES WHATSOEVER RESULTING  FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION,   ARISING OUT OF OR IN    CONNECTION WITH THE USE   OR
# PERFORMANCE OF THIS SOFTWARE.
#
#                                            Anthony Mallet on Fri Mar 23 2012
#

namespace eval server {
  variable sock		{}
  variable channels	[dict create]


  # --- start --------------------------------------------------------------
  #
  proc start { port } {
    variable sock
    set sock [socket -server [namespace code accept] $port]
  }
  namespace export start


  # --- accept -------------------------------------------------------------
  #
  proc accept { channel addr port } {
    variable channels
    infox "connection from $addr:$port on $channel"

    dict set channels $channel [dict create state idle]

    fconfigure $channel -buffering full
    fconfigure $channel -blocking 0
    fileevent $channel readable [list [namespace code eval] $channel]
  }


  # --- terminate ----------------------------------------------------------
  #
  proc terminate { channel } {
    variable channels

    infox "$channel: disconnecting"
    session::clean $channel
    close $channel
    dict unset channels $channel
  }
  namespace export terminate


  # --- eval ---------------------------------------------------------------
  #
  proc eval { channel } {
    variable channels

    # minimalistic http/1.1 state machine
    switch [dict get $channels $channel state] {
      idle {
        if {[gets $channel cmd] < 0} {
          if {[eof $channel]} { terminate $channel }
          return
        }
        infox "$channel: received \"$cmd\"" 1

        switch -regexp -nocase $cmd {
          "^$" {}

          "^(get|post) " {
            set uri [uri decode-args [lindex $cmd 1] args]

            dict set channels $channel method [string tolower [lindex $cmd 0]]
            dict set channels $channel uri $uri
            dict set channels $channel args $args
            dict set channels $channel headers {}
            dict set channels $channel state headers
          }

          default {
            dict set channels $channel method get
            dict set channels $channel uri /
            dict set channels $channel args {}
            dict set channels $channel state headers
          }
        }
      }

      headers {
        if {[gets $channel cmd] < 0} {
          if {[eof $channel]} { terminate $channel }
          return
        }
        infox "$channel: received \"$cmd\"" 1

        if {$cmd == ""} {
          dict set channels $channel body {}
          if {[dict exists $channels $channel headers content-length]} {
            infox "$channel: content-length header, now expecting body" 1
            dict set channels $channel state body
          } else {
            infox "$channel: processing request on $channel" 1
            dispatch $channel [dict get $channels $channel]
            dict set channels $channel state idle
          }
        } else {
          set kv [string first : $cmd]
          if {$kv > 0} {
            set k [string tolower [string trim [string range $cmd 0 $kv-1]]]
            set v [string trim [string range $cmd $kv+1 end]]
            infox "$channel: header \"$k\", value \"$v\"" 1
            dict set channels $channel headers $k $v
          }
        }
      }

      body {
        set body [dict get $channels $channel body]
        set got [string bytelength $body]
        set expect [dict get $channels $channel headers content-length]

        if {$got < $expect} {
          set cmd [read $channel [expr {$expect-$got}]]
          append body $cmd
          dict set channels $channel body $body
          set got [string bytelength $body]
        }

        if {$got >= $expect} {
          infox "$channel: processing request on $channel" 1
          dispatch $channel [dict get $channels $channel]
          dict set channels $channel state idle
        } else {
          infox "$channel: got $got bytes out of $expect" 1
        }
      }
    }

    if {[eof $channel]} { terminate $channel }
    return ""
  }


  # --- ignore -------------------------------------------------------------
  #
  proc ignore { channel } {
    read $channel
    if {[eof $channel]} { terminate $channel }
    return
  }


  # --- reply --------------------------------------------------------------
  #
  proc reply { channel status msg } {
    switch -glob [expr {$status}] {
      2[0-9][0-9] { infox "$channel: reply status $status: $msg" }
      default {
        warnx "$channel: reply status $status: $msg"
        if {[string index $msg 0] != "\{"} {
          set msg \"[string map {\\ \\\\ \" \\\"} $msg]\"
        }
      }
    }

    switch [expr {$status}] {
      200 { set status "200 OK" }
      201 { set status "201 Created" }
      202 { set status "202 Accepted" }
      400 { set status "400 Bad Request" }
      404 { set status "404 Not Found" }
    }

    set len [string bytelength "$msg"]
    set r [catch {
      puts $channel "HTTP/1.1 $status"
      puts $channel "Content-Type: text/plain"
      puts $channel "Content-Length: $len"
      puts $channel ""
      puts -nonewline $channel $msg
      flush $channel
    } m]
    if {$r} {
      warnx "$channel: $m"
      terminate $channel
    }
  }


  # --- stream -------------------------------------------------------------
  #
  proc stream { channel } {
    set r [catch {
      infox "$channel: set to streaming mode" 1

      fileevent $channel readable [list [namespace code ignore] $channel]
      puts $channel "HTTP/1.1 200 OK"
      puts $channel "Content-Type: text/event-stream"
      puts $channel ""
      flush $channel
    } m]
    if {$r} {
      warnx "$channel: $m"
      terminate $channel
    }
  }
  namespace export stream


  # --- dispatch -----------------------------------------------------------
  #
  proc dispatch { channel request } {
    set method [dict get $request method]
    set uri [dict get $request uri]
    set body [dict get $request body]

    catch {dict get $request headers content-type} ctype
    if {$ctype == "application/x-www-form-urlencoded"} {
      uri decode-form $body args
      dict set request args [dict merge [dict get $request args] $args]
    }
    set args [dict get $request args]

    set log "$channel: $method $uri"
    dict for {k v} $args { append log " $k=$v" }
    infox $log

    switch -regexp $uri {
      {^/session/} {
        set e [catch {
          # manage sessions
          set c [lindex [file split $uri] 2]
          set r [session $c $channel $args]
          if {$c != "listen"} { reply $channel 200 $r }
        } r]
        if {$e} { reply $channel 400 $r }
      }

      {^/rpath/} {
        set e [catch {
          # add a load path
          set p [string replace $uri 0 6]
          set r [genomix rpath $p $args]
          reply $channel 200 $r
        } r]
        if {$e} { reply $channel 404 $r }
      }

      {^/load/} {
        set e [catch {
          # load module
          set p [string replace $uri 0 5]
          set r [genomix load $p [dict get $args argv]]
          reply $channel 201 $r
        } r]
        if {$e} { reply $channel 404 $r }
      }

      {^/module/} {
        set r [file split $uri]
        set m [lindex $r 2]
        set c [lindex $r 3]
        if {[info commands ::module::$m] == ""} {
          reply $channel 404 $m
          return
        }
        set e [catch {
          switch $c {
            unload {
              session::broadcast unload $m
              request::clean-all $m $args
              rename ::module::$m {}
              set r "unloaded $m"
            }
            send {
              if {[catch {dict get $args input} input]} {
                set input "{}"
              }
              set r [module::$m send [lindex $r 4] $input]
              request::new $m $r $args
            }
            read {
              set p [lindex $r 4]
              set n [lindex $r 5]
              set r [module::$m read $p $n]
            }
            status {
              set id [lindex $r 4]
              set r [request::status $m $id]
            }
            clean {
              set id [lindex $r 4]
              set r [request::clean $m $id $args]
            }
            default {
              set r [::module::$m $c]
            }
          }
          reply $channel 200 $r
        } r]
        if {$e} { reply $channel 400 $r }
      }

      default {
        reply $channel 400 "Bad request: \"$uri\""
      }
    }
  }


  # create 'server' command
  namespace ensemble create
}
